package matrixbot

import (
	"fmt"
	"io"
	"log"
	"net/http"
	"net/url"
	"strings"
	"time"
)

type MatrixBot struct {
	Token    *MatrixToken
	PushUrl  *url.URL
	handlers map[string]func(bot *MatrixBot, notification Notification) error
}

func NewMatrixBot(pushUrl *url.URL) *MatrixBot {
	return &MatrixBot{
		Token:    nil,
		PushUrl:  pushUrl,
		handlers: make(map[string]func(bot *MatrixBot, notification Notification) error),
	}
}

func (bot *MatrixBot) RefreshToken() error {
	if bot.Token == nil {
		return fmt.Errorf("no refresh token available")
	}
	userData, err := Refresh(bot.Token.RefreshToken)
	if err != nil {
		return err
	}
	bot.Token = &MatrixToken{
		AccessToken:  userData.AccessToken,
		Expires:      time.Now().Add(time.Duration(userData.ExpiresInMs) / 2 * time.Millisecond),
		RefreshToken: userData.RefreshToken,
	}
	return nil
}

func (bot *MatrixBot) Login(username string, password string, deviceId string) error {
	userData, err := Login(username, password, deviceId)
	if err != nil {
		return err
	}
	bot.Token = &MatrixToken{
		AccessToken:  userData.AccessToken,
		Expires:      time.Now().Add(time.Duration(userData.ExpiresInMs) / 2 * time.Millisecond),
		RefreshToken: userData.RefreshToken,
	}
	return nil
}

func (bot *MatrixBot) RefreshTask() {
	for true {
		if bot.Token != nil && time.Now().After(bot.Token.Expires) {
			if err := bot.RefreshToken(); err != nil {
				log.Printf("error refresh token: %s\n", err.Error())
			}
		}
		time.Sleep(1 * time.Second)
	}
}

func (bot *MatrixBot) RegisterPusher() error {
	return SetPusher(*bot.Token, bot.PushUrl.String())
}

func (bot *MatrixBot) HandleFunc(command string, handler func(bot *MatrixBot, notification Notification) error) {
	bot.handlers[command] = handler
}

func (bot *MatrixBot) Serve(endpoint string) {
	http.HandleFunc("/healthz", func(writer http.ResponseWriter, request *http.Request) {
		_, _ = io.WriteString(writer, "OK\n")
	})
	log.Printf("listening for push notifications on %s\n", bot.PushUrl.Path)
	http.HandleFunc(bot.PushUrl.Path, func(writer http.ResponseWriter, request *http.Request) {
		log.Println("Received push notification")
		notification, err := ParseNotification(request.Body)
		if err != nil {
			log.Println(err.Error())
			return
		}
		if notification.EventId == "" {
			return
		}
		err = SetReadReceipt(*bot.Token, notification.RoomId, notification.EventId)
		if err != nil {
			log.Println(err.Error())
			return
		}
		command := strings.SplitN(notification.Content.Body, " ", 2)[0]
		handler, ok := bot.handlers[command]
		if !ok {
			log.Printf("could not find handler for '%s'\n", command)
			return
		}
		err = handler(bot, notification)
		if err != nil {
			log.Println(err.Error())
			return
		}
		writer.WriteHeader(204)
	})
	fmt.Printf("listening for requests on %s\n", endpoint)
	log.Fatal(http.ListenAndServe(endpoint, nil))
}
